@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'test/unit'
 
 class TestRefinement < Test::Unit::TestCase
@@ -227,20 +228,20 @@ class TestRefinement < Test::Unit::TestCase
     assert_equal("Foo#x", foo.x)
   end
 
-  module FixnumSlashExt
-    refine Fixnum do
+  module IntegerSlashExt
+    refine Integer do
       def /(other) quo(other) end
     end
   end
 
   def test_override_builtin_method
     assert_equal(0, 1 / 2)
-    assert_equal(Rational(1, 2), eval_using(FixnumSlashExt, "1 / 2"))
+    assert_equal(Rational(1, 2), eval_using(IntegerSlashExt, "1 / 2"))
     assert_equal(0, 1 / 2)
   end
 
-  module FixnumPlusExt
-    refine Fixnum do
+  module IntegerPlusExt
+    refine Integer do
       def self.method_added(*args); end
       def +(other) "overridden" end
     end
@@ -248,7 +249,7 @@ class TestRefinement < Test::Unit::TestCase
 
   def test_override_builtin_method_with_method_added
     assert_equal(3, 1 + 2)
-    assert_equal("overridden", eval_using(FixnumPlusExt, "1 + 2"))
+    assert_equal("overridden", eval_using(IntegerPlusExt, "1 + 2"))
     assert_equal(3, 1 + 2)
   end
 
@@ -264,10 +265,10 @@ class TestRefinement < Test::Unit::TestCase
   end
 
   module RefineSameClass
-    REFINEMENT1 = refine(Fixnum) {
+    REFINEMENT1 = refine(Integer) {
       def foo; return "foo" end
     }
-    REFINEMENT2 = refine(Fixnum) {
+    REFINEMENT2 = refine(Integer) {
       def bar; return "bar" end
     }
     REFINEMENT3 = refine(String) {
@@ -282,15 +283,15 @@ class TestRefinement < Test::Unit::TestCase
     assert_not_equal(RefineSameClass::REFINEMENT1, RefineSameClass::REFINEMENT3)
   end
 
-  module FixnumFooExt
-    refine Fixnum do
+  module IntegerFooExt
+    refine Integer do
       def foo; "foo"; end
     end
   end
 
   def test_respond_to_should_not_use_refinements
     assert_equal(false, 1.respond_to?(:foo))
-    assert_equal(false, eval_using(FixnumFooExt, "1.respond_to?(:foo)"))
+    assert_equal(false, eval_using(IntegerFooExt, "1.respond_to?(:foo)"))
   end
 
   module StringCmpExt
@@ -381,7 +382,7 @@ class TestRefinement < Test::Unit::TestCase
   def test_refine_in_class
     assert_raise(NoMethodError) do
       Class.new {
-        refine Fixnum do
+        refine Integer do
           def foo
             "c"
           end
@@ -445,13 +446,13 @@ class TestRefinement < Test::Unit::TestCase
 
   module Inspect
     module M
-      Fixnum = refine(Fixnum) {}
+      Integer = refine(Integer) {}
     end
   end
 
   def test_inspect
-    assert_equal("#<refinement:Fixnum@TestRefinement::Inspect::M>",
-                 Inspect::M::Fixnum.inspect)
+    assert_equal("#<refinement:Integer@TestRefinement::Inspect::M>",
+                 Inspect::M::Integer.inspect)
   end
 
   def test_using_method_cache
@@ -502,8 +503,10 @@ class TestRefinement < Test::Unit::TestCase
     end
 
     class C
-      def foo
-        "redefined"
+      EnvUtil.suppress_warning do
+        def foo
+          "redefined"
+        end
       end
     end
   end
@@ -1495,7 +1498,6 @@ class TestRefinement < Test::Unit::TestCase
   end
 
   def test_reopen_refinement_module
-    flag = false
     assert_separately([], <<-"end;")
       $VERBOSE = nil
       class C
@@ -1522,11 +1524,102 @@ class TestRefinement < Test::Unit::TestCase
 
       assert_equal(:bar, C.new.m, "[ruby-core:71423] [Bug #11672]")
     end;
-    flag = true
-  rescue MiniTest::Assertion
-    skip 'expected to fail'
-  ensure
-    raise MiniTest::Assertion, 'this test is expected to fail' if flag
+  end
+
+  module MixedUsing1
+    class C
+      def foo
+        :orig_foo
+      end
+    end
+
+    module R1
+      refine C do
+        def foo
+          [:R1, super]
+        end
+      end
+    end
+
+    module_function
+
+    def foo
+      [:foo, C.new.foo]
+    end
+
+    using R1
+
+    def bar
+      [:bar, C.new.foo]
+    end
+  end
+
+  module MixedUsing2
+    class C
+      def foo
+        :orig_foo
+      end
+    end
+
+    module R1
+      refine C do
+        def foo
+          [:R1_foo, super]
+        end
+      end
+    end
+
+    module R2
+      refine C do
+        def bar
+          [:R2_bar, C.new.foo]
+        end
+
+        using R1
+
+        def baz
+          [:R2_baz, C.new.foo]
+        end
+      end
+    end
+
+    using R2
+    module_function
+    def f1; C.new.bar; end
+    def f2; C.new.baz; end
+  end
+
+  def test_mixed_using
+    assert_equal([:foo, :orig_foo], MixedUsing1.foo)
+    assert_equal([:bar, [:R1, :orig_foo]], MixedUsing1.bar)
+
+    assert_equal([:R2_bar, :orig_foo], MixedUsing2.f1)
+    assert_equal([:R2_baz, [:R1_foo, :orig_foo]], MixedUsing2.f2)
+  end
+
+  module MethodMissing
+    class Foo
+    end
+
+    module Bar
+      refine Foo  do
+        def method_missing(mid, *args)
+          "method_missing refined"
+        end
+      end
+    end
+
+    using Bar
+
+    def self.call_undefined_method
+      Foo.new.foo
+    end
+  end
+
+  def test_method_missing
+    assert_raise(NoMethodError) do
+      MethodMissing.call_undefined_method
+    end
   end
 
   private
