@@ -1,13 +1,31 @@
-#define RUBY_VERSION "2.3.0"
-#define RUBY_RELEASE_DATE "2015-11-11"
+#define RUBY_VERSION "2.4.0"
+#define RUBY_RELEASE_DATE RUBY_RELEASE_YEAR_STR"-"RUBY_RELEASE_MONTH_STR"-"RUBY_RELEASE_DAY_STR
 #define RUBY_PATCHLEVEL -1
 
-#define RUBY_RELEASE_YEAR 2015
-#define RUBY_RELEASE_MONTH 11
-#define RUBY_RELEASE_DAY 11
+#define RUBY_RELEASE_YEAR 2016
+#define RUBY_RELEASE_MONTH 6
+#define RUBY_RELEASE_DAY 20
 
 #include "ruby/version.h"
 
+#ifndef TOKEN_PASTE
+#define TOKEN_PASTE(x,y) x##y
+#endif
+#define ONLY_ONE_DIGIT(x) TOKEN_PASTE(10,x) < 1000
+#define WITH_ZERO_PADDING(x) TOKEN_PASTE(0,x)
+#define RUBY_BIRTH_YEAR_STR STRINGIZE(RUBY_BIRTH_YEAR)
+#define RUBY_RELEASE_YEAR_STR STRINGIZE(RUBY_RELEASE_YEAR)
+#if ONLY_ONE_DIGIT(RUBY_RELEASE_MONTH)
+#define RUBY_RELEASE_MONTH_STR STRINGIZE(WITH_ZERO_PADDING(RUBY_RELEASE_MONTH))
+#else
+#define RUBY_RELEASE_MONTH_STR STRINGIZE(RUBY_RELEASE_MONTH)
+#endif
+#if ONLY_ONE_DIGIT(RUBY_RELEASE_DAY)
+#define RUBY_RELEASE_DAY_STR STRINGIZE(WITH_ZERO_PADDING(RUBY_RELEASE_DAY))
+#else
+#define RUBY_RELEASE_DAY_STR STRINGIZE(RUBY_RELEASE_DAY)
+#endif
+
 #if !defined RUBY_LIB_VERSION && defined RUBY_LIB_VERSION_STYLE
 # if RUBY_LIB_VERSION_STYLE == 3
 #   define RUBY_LIB_VERSION STRINGIZE(RUBY_API_VERSION_MAJOR)"."STRINGIZE(RUBY_API_VERSION_MINOR)"."STRINGIZE(RUBY_API_VERSION_TEENY)
@@ -50,6 +68,6 @@
     "["RUBY_PLATFORM"]"
 # define RUBY_COPYRIGHT		    \
     "ruby - Copyright (C) "	    \
-    STRINGIZE(RUBY_BIRTH_YEAR)"-"   \
-    STRINGIZE(RUBY_RELEASE_YEAR)" " \
+    RUBY_BIRTH_YEAR_STR"-"   \
+    RUBY_RELEASE_YEAR_STR" " \
     RUBY_AUTHOR
