@@ -2,7 +2,7 @@
 
   vm.c -
 
-  $Author: nobu $
+  $Author: shugo $
 
   Copyright (C) 2004-2007 Koichi Sasada
 
@@ -23,8 +23,9 @@
 #include "probes.h"
 #include "probes_helper.h"
 
-VALUE rb_str_append_literal(VALUE str, VALUE str2);
+VALUE rb_str_concat_literals(size_t, const VALUE*);
 
+PUREFUNC(static inline VALUE *VM_EP_LEP(VALUE *));
 static inline VALUE *
 VM_EP_LEP(VALUE *ep)
 {
@@ -60,6 +61,7 @@ rb_vm_ep_local_ep(VALUE *ep)
     return VM_EP_LEP(ep);
 }
 
+PUREFUNC(static inline VALUE *VM_CF_LEP(const rb_control_frame_t * const cfp));
 static inline VALUE *
 VM_CF_LEP(const rb_control_frame_t * const cfp)
 {
@@ -72,6 +74,7 @@ VM_CF_PREV_EP(const rb_control_frame_t * const cfp)
     return VM_EP_PREV_EP(cfp->ep);
 }
 
+PUREFUNC(static inline rb_block_t *VM_CF_BLOCK_PTR(const rb_control_frame_t * const cfp));
 static inline rb_block_t *
 VM_CF_BLOCK_PTR(const rb_control_frame_t * const cfp)
 {
@@ -86,25 +89,76 @@ rb_vm_control_frame_block_ptr(const rb_control_frame_t *cfp)
 }
 
 static rb_cref_t *
-vm_cref_new(VALUE klass, rb_method_visibility_t visi, const rb_cref_t *prev_cref)
+vm_cref_new0(VALUE klass, rb_method_visibility_t visi, int module_func, rb_cref_t *prev_cref, int pushed_by_eval, int use_prev_prev)
 {
+    VALUE refinements = Qnil;
+    int omod_shared = FALSE;
+    rb_cref_t *cref;
+
+    /* scope */
     union {
 	rb_scope_visibility_t visi;
 	VALUE value;
     } scope_visi;
-    scope_visi.visi.method_visi = visi;
-    scope_visi.visi.module_func = 0;
 
-    return (rb_cref_t *)rb_imemo_new(imemo_cref, klass, (VALUE)prev_cref, scope_visi.value, Qnil);
+    scope_visi.visi.method_visi = visi;
+    scope_visi.visi.module_func = module_func;
+
+    /* refinements */
+    if (prev_cref != NULL && prev_cref != (void *)1 /* TODO: why CREF_NEXT(cref) is 1? */) {
+	refinements = CREF_REFINEMENTS(prev_cref);
+
+	if (!NIL_P(refinements)) {
+	    omod_shared = TRUE;
+	    CREF_OMOD_SHARED_SET(prev_cref);
+	}
+    }
+
+    cref = (rb_cref_t *)rb_imemo_new(imemo_cref, klass, (VALUE)(use_prev_prev ? CREF_NEXT(prev_cref) : prev_cref), scope_visi.value, refinements);
+
+    if (pushed_by_eval) CREF_PUSHED_BY_EVAL_SET(cref);
+    if (omod_shared) CREF_OMOD_SHARED_SET(cref);
+
+    return cref;
+}
+
+static rb_cref_t *
+vm_cref_new(VALUE klass, rb_method_visibility_t visi, int module_func, rb_cref_t *prev_cref, int pushed_by_eval)
+{
+    return vm_cref_new0(klass, visi, module_func, prev_cref, pushed_by_eval, FALSE);
+}
+
+static rb_cref_t *
+vm_cref_new_use_prev(VALUE klass, rb_method_visibility_t visi, int module_func, rb_cref_t *prev_cref, int pushed_by_eval)
+{
+    return vm_cref_new0(klass, visi, module_func, prev_cref, pushed_by_eval, TRUE);
+}
+
+static rb_cref_t *
+vm_cref_dup(const rb_cref_t *cref)
+{
+    VALUE klass = CREF_CLASS(cref);
+    const rb_scope_visibility_t *visi = CREF_SCOPE_VISI(cref);
+    rb_cref_t *next_cref = CREF_NEXT(cref), *new_cref;
+    int pushed_by_eval = CREF_PUSHED_BY_EVAL(cref);
+
+    new_cref = vm_cref_new(klass, visi->method_visi, visi->module_func, next_cref, pushed_by_eval);
+
+    if (!NIL_P(CREF_REFINEMENTS(cref))) {
+	CREF_REFINEMENTS_SET(new_cref, rb_hash_dup(CREF_REFINEMENTS(cref)));
+	CREF_OMOD_SHARED_UNSET(new_cref);
+    }
+
+    return new_cref;
 }
 
 static rb_cref_t *
 vm_cref_new_toplevel(rb_thread_t *th)
 {
-    rb_cref_t *cref = vm_cref_new(rb_cObject, METHOD_VISI_PRIVATE /* toplevel visibility is private */, NULL);
+    rb_cref_t *cref = vm_cref_new(rb_cObject, METHOD_VISI_PRIVATE /* toplevel visibility is private */, FALSE, NULL, FALSE);
 
     if (th->top_wrapper) {
-	cref = vm_cref_new(th->top_wrapper, METHOD_VISI_PRIVATE, cref);
+	cref = vm_cref_new(th->top_wrapper, METHOD_VISI_PRIVATE, FALSE, cref, FALSE);
     }
 
     return cref;
@@ -408,8 +462,7 @@ rb_vm_rewind_cfp(rb_thread_t *th, rb_control_frame_t *cfp)
 void
 rb_frame_pop(void)
 {
-    rb_thread_t *th = GET_THREAD();
-    vm_pop_frame(th);
+    ONLY_FOR_INTERNAL_USE("rb_frame_pop()");
 }
 
 /* at exit */
@@ -417,20 +470,25 @@ rb_frame_pop(void)
 void
 ruby_vm_at_exit(void (*func)(rb_vm_t *))
 {
-    rb_ary_push((VALUE)&GET_VM()->at_exit, (VALUE)func);
+    rb_vm_t *vm = GET_VM();
+    rb_at_exit_list *nl = ALLOC(rb_at_exit_list);
+    nl->func = func;
+    nl->next = vm->at_exit;
+    vm->at_exit = nl;
 }
 
 static void
 ruby_vm_run_at_exit_hooks(rb_vm_t *vm)
 {
-    VALUE hook = (VALUE)&vm->at_exit;
+    rb_at_exit_list *l = vm->at_exit;
 
-    while (RARRAY_LEN(hook) > 0) {
-	typedef void rb_vm_at_exit_func(rb_vm_t*);
-	rb_vm_at_exit_func *func = (rb_vm_at_exit_func*)rb_ary_pop(hook);
+    while (l) {
+	rb_at_exit_list* t = l->next;
+	rb_vm_at_exit_func *func = l->func;
+	free(l);
+	l = t;
 	(*func)(vm);
     }
-    rb_ary_free(hook);
 }
 
 /* Env */
@@ -508,16 +566,21 @@ check_env_value(VALUE envval)
     return Qnil;		/* unreachable */
 }
 
-/* return Qfalse if proc was already created */
-static VALUE
-vm_make_proc_from_block(rb_thread_t *th, rb_block_t *block)
+/* return FALSE if proc was already created */
+static int
+vm_make_proc_from_block(rb_thread_t *th, rb_block_t *block, VALUE *procptr)
 {
     if (!block->proc) {
-	block->proc = rb_vm_make_proc(th, block, rb_cProc);
-	return block->proc;
+	*procptr = block->proc = rb_vm_make_proc(th, block, rb_cProc);
+	return TRUE;
+    }
+    else if (SYMBOL_P(block->proc)) {
+	*procptr = rb_sym_to_proc(block->proc);
+	return TRUE;
     }
     else {
-	return Qfalse;
+	*procptr = block->proc;
+	return FALSE;
     }
 }
 
@@ -552,7 +615,7 @@ vm_make_env_each(rb_thread_t *const th, rb_control_frame_t *const cfp)
     else {
 	rb_block_t *block = VM_EP_BLOCK_PTR(ep);
 
-	if (block && (blockprocval = vm_make_proc_from_block(th, block)) != Qfalse) {
+	if (block && vm_make_proc_from_block(th, block, &blockprocval)) {
 	    rb_proc_t *p;
 	    GetProcPtr(blockprocval, p);
 	    *ep = VM_ENVVAL_BLOCK_PTR(&p->block);
@@ -704,6 +767,17 @@ rb_vm_env_local_variables(const rb_env_t *env)
     return local_var_list_finish(&vars);
 }
 
+VALUE
+rb_iseq_local_variables(const rb_iseq_t *iseq)
+{
+    struct local_var_list vars;
+    local_var_list_init(&vars);
+    while (collect_local_variables_in_iseq(iseq, &vars)) {
+	iseq = iseq->body->parent_iseq;
+    }
+    return local_var_list_finish(&vars);
+}
+
 /* Proc */
 
 static inline VALUE
@@ -818,7 +892,7 @@ rb_binding_add_dynavars(rb_binding_t *bind, int dyncount, const ID *dynvars)
 	iseq = rb_iseq_new(node, base_iseq->body->location.label, path, path, base_iseq, ISEQ_TYPE_EVAL);
     }
     else {
-	VALUE tempstr = rb_str_new2("<temp>");
+	VALUE tempstr = rb_fstring_cstr("<temp>");
 	iseq = rb_iseq_new_top(node, tempstr, tempstr, tempstr, NULL);
     }
     node->u1.tbl = 0; /* reset table */
@@ -850,15 +924,12 @@ invoke_block(rb_thread_t *th, const rb_iseq_t *iseq, VALUE self, const rb_block_
 }
 
 static VALUE
-invoke_bmethod(rb_thread_t *th, const rb_iseq_t *iseq, VALUE self, const rb_block_t *block, int type, int opt_pc)
+invoke_bmethod(rb_thread_t *th, const rb_iseq_t *iseq, VALUE self, const rb_block_t *block, const rb_callable_method_entry_t *me, int type, int opt_pc)
 {
     /* bmethod */
     int arg_size = iseq->body->param.size;
-    const rb_callable_method_entry_t *me = th->passed_bmethod_me;
     VALUE ret;
 
-    th->passed_bmethod_me = NULL;
-
     vm_push_frame(th, iseq, type | VM_FRAME_FLAG_FINISH | VM_FRAME_FLAG_BMETHOD, self,
 		  VM_ENVVAL_PREV_EP_PTR(block->ep),
 		  (VALUE)me, /* cref or method (TODO: can we ignore cref?) */
@@ -883,10 +954,13 @@ invoke_block_from_c_0(rb_thread_t *th, const rb_block_t *block,
 	return Qnil;
     }
     else if (LIKELY(RUBY_VM_NORMAL_ISEQ_P(block->iseq))) {
-	const rb_iseq_t *iseq = block->iseq;
+	const rb_iseq_t *iseq = rb_iseq_check(block->iseq);
 	int i, opt_pc;
 	int type = block_proc_is_lambda(block->proc) ? VM_FRAME_MAGIC_LAMBDA : VM_FRAME_MAGIC_BLOCK;
 	VALUE *sp = th->cfp->sp;
+	const rb_callable_method_entry_t *me = th->passed_bmethod_me;
+
+	th->passed_bmethod_me = NULL;
 
 	for (i=0; i<argc; i++) {
 	    sp[i] = argv[i];
@@ -895,11 +969,11 @@ invoke_block_from_c_0(rb_thread_t *th, const rb_block_t *block,
 	opt_pc = vm_yield_setup_args(th, iseq, argc, sp, blockptr,
 				     (type == VM_FRAME_MAGIC_LAMBDA ? (splattable ? arg_setup_lambda : arg_setup_method) : arg_setup_block));
 
-	if (th->passed_bmethod_me == NULL) {
+	if (me == NULL) {
 	    return invoke_block(th, iseq, self, block, cref, type, opt_pc);
 	}
 	else {
-	    return invoke_bmethod(th, iseq, self, block, type, opt_pc);
+	    return invoke_bmethod(th, iseq, self, block, me, type, opt_pc);
 	}
 
     }
@@ -976,7 +1050,7 @@ vm_invoke_proc(rb_thread_t *th, rb_proc_t *proc, VALUE self,
     th->safe_level = stored_safe;
 
     if (state) {
-	JUMP_TAG(state);
+	TH_JUMP_TAG(th, state);
     }
     return val;
 }
@@ -1148,6 +1222,15 @@ rb_vm_cref(void)
     return rb_vm_get_cref(cfp->ep);
 }
 
+rb_cref_t *
+rb_vm_cref_replace_with_duplicated_cref(void)
+{
+    rb_thread_t *th = GET_THREAD();
+    rb_control_frame_t *cfp = rb_vm_get_ruby_level_next_cfp(th, th->cfp);
+    rb_cref_t *cref = vm_cref_replace_with_duplicated_cref(cfp->ep);
+    return cref;
+}
+
 const rb_cref_t *
 rb_vm_cref_in_context(VALUE self, VALUE cbase)
 {
@@ -1155,6 +1238,7 @@ rb_vm_cref_in_context(VALUE self, VALUE cbase)
     const rb_control_frame_t *cfp = rb_vm_get_ruby_level_next_cfp(th, th->cfp);
     const rb_cref_t *cref;
     if (cfp->self != self) return NULL;
+    if (!vm_env_cref_by_cref(cfp->ep)) return NULL;
     cref = rb_vm_get_cref(cfp->ep);
     if (CREF_CLASS(cref) != cbase) return NULL;
     return cref;
@@ -1304,15 +1388,17 @@ static st_table *vm_opt_method_table = 0;
 static int
 vm_redefinition_check_flag(VALUE klass)
 {
-    if (klass == rb_cFixnum) return FIXNUM_REDEFINED_OP_FLAG;
+    if (klass == rb_cInteger) return INTEGER_REDEFINED_OP_FLAG;
     if (klass == rb_cFloat)  return FLOAT_REDEFINED_OP_FLAG;
     if (klass == rb_cString) return STRING_REDEFINED_OP_FLAG;
     if (klass == rb_cArray)  return ARRAY_REDEFINED_OP_FLAG;
     if (klass == rb_cHash)   return HASH_REDEFINED_OP_FLAG;
-    if (klass == rb_cBignum) return BIGNUM_REDEFINED_OP_FLAG;
     if (klass == rb_cSymbol) return SYMBOL_REDEFINED_OP_FLAG;
     if (klass == rb_cTime)   return TIME_REDEFINED_OP_FLAG;
     if (klass == rb_cRegexp) return REGEXP_REDEFINED_OP_FLAG;
+    if (klass == rb_cNilClass) return NIL_REDEFINED_OP_FLAG;
+    if (klass == rb_cTrueClass) return TRUE_REDEFINED_OP_FLAG;
+    if (klass == rb_cFalseClass) return FALSE_REDEFINED_OP_FLAG;
     return 0;
 }
 
@@ -1320,6 +1406,9 @@ static void
 rb_vm_check_redefinition_opt_method(const rb_method_entry_t *me, VALUE klass)
 {
     st_data_t bop;
+    if (RB_TYPE_P(klass, T_ICLASS) && FL_TEST(klass, RICLASS_IS_ORIGIN)) {
+       klass = RBASIC_CLASS(klass);
+    }
     if (me->def->type == VM_METHOD_TYPE_CFUNC) {
 	if (st_lookup(vm_opt_method_table, (st_data_t)me, &bop)) {
 	    int flag = vm_redefinition_check_flag(klass);
@@ -1371,26 +1460,29 @@ vm_init_redefined_flag(void)
 
 #define OP(mid_, bop_) (mid = id##mid_, bop = BOP_##bop_, ruby_vm_redefined_flag[bop] = 0)
 #define C(k) add_opt_method(rb_c##k, mid, bop)
-    OP(PLUS, PLUS), (C(Fixnum), C(Float), C(String), C(Array));
-    OP(MINUS, MINUS), (C(Fixnum), C(Float));
-    OP(MULT, MULT), (C(Fixnum), C(Float));
-    OP(DIV, DIV), (C(Fixnum), C(Float));
-    OP(MOD, MOD), (C(Fixnum), C(Float));
-    OP(Eq, EQ), (C(Fixnum), C(Float), C(String));
-    OP(Eqq, EQQ), (C(Fixnum), C(Bignum), C(Float), C(Symbol), C(String));
-    OP(LT, LT), (C(Fixnum), C(Float));
-    OP(LE, LE), (C(Fixnum), C(Float));
-    OP(GT, GT), (C(Fixnum), C(Float));
-    OP(GE, GE), (C(Fixnum), C(Float));
+    OP(PLUS, PLUS), (C(Integer), C(Float), C(String), C(Array));
+    OP(MINUS, MINUS), (C(Integer), C(Float));
+    OP(MULT, MULT), (C(Integer), C(Float));
+    OP(DIV, DIV), (C(Integer), C(Float));
+    OP(MOD, MOD), (C(Integer), C(Float));
+    OP(Eq, EQ), (C(Integer), C(Float), C(String));
+    OP(Eqq, EQQ), (C(Integer), C(Float), C(Symbol), C(String),
+		   C(NilClass), C(TrueClass), C(FalseClass));
+    OP(LT, LT), (C(Integer), C(Float));
+    OP(LE, LE), (C(Integer), C(Float));
+    OP(GT, GT), (C(Integer), C(Float));
+    OP(GE, GE), (C(Integer), C(Float));
     OP(LTLT, LTLT), (C(String), C(Array));
     OP(AREF, AREF), (C(Array), C(Hash));
     OP(ASET, ASET), (C(Array), C(Hash));
     OP(Length, LENGTH), (C(Array), C(String), C(Hash));
     OP(Size, SIZE), (C(Array), C(String), C(Hash));
     OP(EmptyP, EMPTY_P), (C(Array), C(String), C(Hash));
-    OP(Succ, SUCC), (C(Fixnum), C(String), C(Time));
+    OP(Succ, SUCC), (C(Integer), C(String), C(Time));
     OP(EqTilde, MATCH), (C(Regexp), C(String));
     OP(Freeze, FREEZE), (C(String));
+    OP(Max, MAX), (C(Array));
+    OP(Min, MIN), (C(Array));
 #undef C
 #undef OP
 }
@@ -1586,7 +1678,7 @@ vm_exec(rb_thread_t *th)
 		EXEC_EVENT_HOOK(th, RUBY_EVENT_C_RETURN, th->cfp->self,
 				rb_vm_frame_method_entry(th->cfp)->called_id,
 				rb_vm_frame_method_entry(th->cfp)->owner, Qnil);
-		RUBY_DTRACE_METHOD_RETURN_HOOK(th,
+		RUBY_DTRACE_CMETHOD_RETURN_HOOK(th,
 					       rb_vm_frame_method_entry(th->cfp)->owner,
 					       rb_vm_frame_method_entry(th->cfp)->called_id);
 	    }
@@ -1741,6 +1833,7 @@ vm_exec(rb_thread_t *th)
 
 	if (catch_iseq != NULL) { /* found catch table */
 	    /* enter catch scope */
+	    rb_iseq_check(catch_iseq);
 	    cfp->sp = vm_base_ptr(cfp) + cont_sp;
 	    cfp->pc = cfp->iseq->body->iseq_encoded + cont_pc;
 
@@ -1768,7 +1861,7 @@ vm_exec(rb_thread_t *th)
 		vm_pop_frame(th);
 		th->errinfo = (VALUE)err;
 		TH_TMPPOP_TAG();
-		JUMP_TAG(state);
+		TH_JUMP_TAG(th, state);
 	    }
 	    else {
 		th->cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(th->cfp);
@@ -1941,6 +2034,13 @@ rb_vm_add_root_module(ID id, VALUE module)
     return TRUE;
 }
 
+static int
+free_loading_table_entry(st_data_t key, st_data_t value, st_data_t arg)
+{
+    xfree((char *)key);
+    return ST_DELETE;
+}
+
 int
 ruby_vm_destruct(rb_vm_t *vm)
 {
@@ -1956,6 +2056,15 @@ ruby_vm_destruct(rb_vm_t *vm)
 	}
 	rb_vm_living_threads_init(vm);
 	ruby_vm_run_at_exit_hooks(vm);
+	if (vm->loading_table) {
+	    st_foreach(vm->loading_table, free_loading_table_entry, 0);
+	    st_free_table(vm->loading_table);
+	    vm->loading_table = 0;
+	}
+	if (vm->frozen_strings) {
+	    st_free_table(vm->frozen_strings);
+	    vm->frozen_strings = 0;
+	}
 	rb_vm_gvl_destroy(vm);
 	if (objspace) {
 	    rb_objspace_free(objspace);
@@ -1971,20 +2080,15 @@ ruby_vm_destruct(rb_vm_t *vm)
 static size_t
 vm_memsize(const void *ptr)
 {
-    if (ptr) {
-	const rb_vm_t *vmobj = ptr;
-	size_t size = sizeof(rb_vm_t);
+    const rb_vm_t *vmobj = ptr;
+    size_t size = sizeof(rb_vm_t);
 
-	size += vmobj->living_thread_num * sizeof(rb_thread_t);
+    size += vmobj->living_thread_num * sizeof(rb_thread_t);
 
-	if (vmobj->defined_strings) {
-	    size += DEFINED_EXPR * sizeof(VALUE);
-	}
-	return size;
-    }
-    else {
-	return 0;
+    if (vmobj->defined_strings) {
+	size += DEFINED_EXPR * sizeof(VALUE);
     }
+    return size;
 }
 
 static const rb_data_type_t vm_data_type = {
@@ -2074,8 +2178,6 @@ vm_init2(rb_vm_t *vm)
     MEMZERO(vm, rb_vm_t, 1);
     rb_vm_living_threads_init(vm);
     vm->src_encoding_index = -1;
-    vm->at_exit.basic.flags = (T_ARRAY | RARRAY_EMBED_FLAG) & ~RARRAY_EMBED_LEN_MASK; /* len set 0 */
-    rb_obj_hide((VALUE)&vm->at_exit);
 
     vm_default_params_setup(vm);
 }
@@ -2222,21 +2324,16 @@ thread_free(void *ptr)
 static size_t
 thread_memsize(const void *ptr)
 {
-    if (ptr) {
-	const rb_thread_t *th = ptr;
-	size_t size = sizeof(rb_thread_t);
+    const rb_thread_t *th = ptr;
+    size_t size = sizeof(rb_thread_t);
 
-	if (!th->root_fiber) {
-	    size += th->stack_size * sizeof(VALUE);
-	}
-	if (th->local_storage) {
-	    size += st_memsize(th->local_storage);
-	}
-	return size;
+    if (!th->root_fiber) {
+	size += th->stack_size * sizeof(VALUE);
     }
-    else {
-	return 0;
+    if (th->local_storage) {
+	size += st_memsize(th->local_storage);
     }
+    return size;
 }
 
 #define thread_data_type ruby_threadptr_data_type
@@ -2337,25 +2434,28 @@ rb_thread_alloc(VALUE klass)
 }
 
 static void
-vm_define_method(rb_thread_t *th, VALUE obj, ID id, VALUE iseqval,
-		 rb_num_t is_singleton, rb_cref_t *cref)
+vm_define_method(rb_thread_t *th, VALUE obj, ID id, VALUE iseqval, int is_singleton)
 {
-    VALUE klass = CREF_CLASS(cref);
-    const rb_scope_visibility_t *scope_visi = CREF_SCOPE_VISI(cref);
-    rb_method_visibility_t visi = scope_visi->method_visi;
+    VALUE klass;
+    rb_method_visibility_t visi;
+    rb_cref_t *cref = rb_vm_cref();
+
+    if (!is_singleton) {
+	klass = CREF_CLASS(cref);
+	visi = rb_scope_visibility_get();
+    }
+    else { /* singleton */
+	klass = rb_singleton_class(obj); /* class and frozen checked in this API */
+	visi = METHOD_VISI_PUBLIC;
+    }
 
     if (NIL_P(klass)) {
 	rb_raise(rb_eTypeError, "no class/module to add method");
     }
 
-    if (is_singleton) {
-	klass = rb_singleton_class(obj); /* class and frozen checked in this API */
-	visi = METHOD_VISI_PUBLIC;
-    }
-
     rb_add_method_iseq(klass, id, (const rb_iseq_t *)iseqval, cref, visi);
 
-    if (!is_singleton && scope_visi->module_func) {
+    if (!is_singleton && rb_scope_module_func_check()) {
 	klass = rb_singleton_class(klass);
 	rb_add_method_iseq(klass, id, (const rb_iseq_t *)iseqval, cref, METHOD_VISI_PUBLIC);
     }
@@ -2371,10 +2471,10 @@ vm_define_method(rb_thread_t *th, VALUE obj, ID id, VALUE iseqval,
 } while (0)
 
 static VALUE
-m_core_define_method(VALUE self, VALUE cbase, VALUE sym, VALUE iseqval)
+m_core_define_method(VALUE self, VALUE sym, VALUE iseqval)
 {
     REWIND_CFP({
-	vm_define_method(GET_THREAD(), cbase, SYM2ID(sym), iseqval, 0, rb_vm_cref());
+	vm_define_method(GET_THREAD(), Qnil, SYM2ID(sym), iseqval, FALSE);
     });
     return sym;
 }
@@ -2383,7 +2483,7 @@ static VALUE
 m_core_define_singleton_method(VALUE self, VALUE cbase, VALUE sym, VALUE iseqval)
 {
     REWIND_CFP({
-	vm_define_method(GET_THREAD(), cbase, SYM2ID(sym), iseqval, 1, rb_vm_cref());
+	vm_define_method(GET_THREAD(), cbase, SYM2ID(sym), iseqval, TRUE);
     });
     return sym;
 }
@@ -2583,7 +2683,7 @@ Init_VM(void)
     rb_define_method_id(klass, id_core_set_method_alias, m_core_set_method_alias, 3);
     rb_define_method_id(klass, id_core_set_variable_alias, m_core_set_variable_alias, 2);
     rb_define_method_id(klass, id_core_undef_method, m_core_undef_method, 2);
-    rb_define_method_id(klass, id_core_define_method, m_core_define_method, 3);
+    rb_define_method_id(klass, id_core_define_method, m_core_define_method, 2);
     rb_define_method_id(klass, id_core_define_singleton_method, m_core_define_singleton_method, 3);
     rb_define_method_id(klass, id_core_set_postexe, m_core_set_postexe, 0);
     rb_define_method_id(klass, id_core_hash_from_ary, m_core_hash_from_ary, 1);
@@ -2756,10 +2856,10 @@ Init_VM(void)
 #if VM_COLLECT_USAGE_DETAILS
     /* ::RubyVM::USAGE_ANALYSIS_* */
 #define define_usage_analysis_hash(name) /* shut up rdoc -C */ \
-    rb_define_const(rb_cRubyVM, "USAGE_ANALYSIS_"#name, rb_hash_new())
-    define_usage_analysis_hash("INSN");
-    define_usage_analysis_hash("REGS");
-    define_usage_analysis_hash("INSN_BIGRAM");
+    rb_define_const(rb_cRubyVM, "USAGE_ANALYSIS_" #name, rb_hash_new())
+    define_usage_analysis_hash(INSN);
+    define_usage_analysis_hash(REGS);
+    define_usage_analysis_hash(INSN_BIGRAM);
 
     rb_define_singleton_method(rb_cRubyVM, "USAGE_ANALYSIS_INSN_STOP", usage_analysis_insn_stop, 0);
     rb_define_singleton_method(rb_cRubyVM, "USAGE_ANALYSIS_OPERAND_STOP", usage_analysis_operand_stop, 0);
@@ -2817,7 +2917,7 @@ Init_VM(void)
     {
 	rb_vm_t *vm = ruby_current_vm;
 	rb_thread_t *th = GET_THREAD();
-	VALUE filename = rb_str_new2("<main>");
+	VALUE filename = rb_fstring_cstr("<main>");
 	const rb_iseq_t *iseq = rb_iseq_new(0, filename, filename, Qnil, 0, ISEQ_TYPE_TOP);
         volatile VALUE th_self;
 
@@ -2841,7 +2941,7 @@ Init_VM(void)
 	th->cfp->pc = iseq->body->iseq_encoded;
 	th->cfp->self = th->top_self;
 
-	th->cfp->ep[-1] = (VALUE)vm_cref_new(rb_cObject, METHOD_VISI_PRIVATE, NULL);
+	th->cfp->ep[-1] = (VALUE)vm_cref_new(rb_cObject, METHOD_VISI_PRIVATE, FALSE, NULL, FALSE);
 
 	/*
 	 * The Binding of the top level scope
@@ -2926,13 +3026,13 @@ Init_top_self(void)
     rb_define_alias(rb_singleton_class(rb_vm_top_self()), "inspect", "to_s");
 }
 
-VALUE *
+static VALUE *
 ruby_vm_verbose_ptr(rb_vm_t *vm)
 {
     return &vm->verbose;
 }
 
-VALUE *
+static VALUE *
 ruby_vm_debug_ptr(rb_vm_t *vm)
 {
     return &vm->debug;
