@@ -691,8 +691,6 @@ setup_parameters_complex(rb_thread_t * const th, const rb_iseq_t * const iseq,
     return opt_pc;
 }
 
-VALUE rb_exc_set_backtrace(VALUE exc, VALUE bt);
-
 static void
 raise_argument_error(rb_thread_t *th, const rb_iseq_t *iseq, const VALUE exc)
 {
@@ -785,11 +783,6 @@ vm_caller_setup_arg_block(const rb_thread_t *th, rb_control_frame_t *reg_cfp,
 	    calling->blockptr->iseq = (rb_iseq_t *)proc;
 	    calling->blockptr->proc = proc;
 	}
-	else if (RUBY_VM_IFUNC_P(proc)) {
-	    calling->blockptr = RUBY_VM_GET_BLOCK_PTR_IN_CFP(reg_cfp);
-	    calling->blockptr->iseq = (rb_iseq_t *)proc;
-	    calling->blockptr->proc = proc;
-	}
 	else {
 	    if (!rb_obj_is_proc(proc)) {
 		VALUE b;
