@@ -333,7 +333,7 @@ location_to_str(rb_backtrace_location_t *loc)
 	}
 	else {
 	    rb_thread_t *th = GET_THREAD();
-	    file = th->vm->progname ? th->vm->progname : ruby_engine_name;
+	    file = th->vm->progname;
 	    lineno = INT2FIX(0);
 	}
 	name = rb_id2str(loc->body.cfunc.mid);
@@ -683,7 +683,7 @@ oldbt_init(void *ptr, size_t dmy)
     struct oldbt_arg *arg = (struct oldbt_arg *)ptr;
     rb_thread_t *th = GET_THREAD();
 
-    arg->filename = th->vm->progname ? th->vm->progname : ruby_engine_name;;
+    arg->filename = th->vm->progname;
     arg->lineno = 0;
 }
 
@@ -1191,7 +1191,7 @@ rb_debug_inspector_open(rb_debug_inspector_func_t func, void *data)
     /* invalidate bindings? */
 
     if (state) {
-	JUMP_TAG(state);
+	TH_JUMP_TAG(th, state);
     }
 
     return result;
@@ -1248,25 +1248,25 @@ rb_profile_frames(int start, int limit, VALUE *buff, int *lines)
     int i;
     rb_thread_t *th = GET_THREAD();
     rb_control_frame_t *cfp = th->cfp, *end_cfp = RUBY_VM_END_CONTROL_FRAME(th);
+    const rb_callable_method_entry_t *cme;
 
     for (i=0; i<limit && cfp != end_cfp;) {
-	const rb_callable_method_entry_t *cme = rb_vm_frame_method_entry(cfp);
-
-	if ((cme && cme->def->type == VM_METHOD_TYPE_ISEQ) || (cfp->iseq && cfp->pc)) {
+	if (cfp->iseq && cfp->pc) {
 	    if (start > 0) {
 		start--;
 		continue;
 	    }
 
 	    /* record frame info */
-	    if (cme) {
+	    cme = rb_vm_frame_method_entry(cfp);
+	    if (cme && cme->def->type == VM_METHOD_TYPE_ISEQ) {
 		buff[i] = (VALUE)cme;
 	    }
 	    else {
 		buff[i] = (VALUE)cfp->iseq;
 	    }
 
-	    if (cfp->iseq && lines) lines[i] = calc_lineno(cfp->iseq, cfp->pc);
+	    if (lines) lines[i] = calc_lineno(cfp->iseq, cfp->pc);
 
 	    i++;
 	}
@@ -1344,12 +1344,12 @@ frame2klass(VALUE frame)
 
     if (RB_TYPE_P(frame, T_IMEMO)) {
 	const rb_callable_method_entry_t *cme = (rb_callable_method_entry_t *)frame;
-	VM_ASSERT(imemo_type(frame) == imemo_ment);
-	return cme->defined_class;
-    }
-    else {
-	return Qnil;
+
+	if (imemo_type(frame) == imemo_ment) {
+	    return cme->defined_class;
+	}
     }
+    return Qnil;
 }
 
 VALUE
